/**
 *	Based on the fact that Object.sc "catches" all
 *	calls to unknown methods in "doesNotUnderstand",
 *	we exploit this behaviour to create an easy wrapper
 *	class for Java object control in SwingOSC.
 *
 *	@version	0.11, 02-Aug-06
 *	@author	Hanns Holger Rutz
 */
JavaObject {
	classvar allObjects;
	var <server, <id;

	*initClass {
		UI.registerForShutdown({ this.destroyAll; });
	}

	*new { arg className, server ... args ;
		^super.new.prInitJavaObject( className, server, args );
	}
	
	prInitJavaObject { arg className, argServer, args;
		var msg;
		
		server		= argServer ?? SwingOSC.default;
		allObjects	= allObjects.add( this );	// array grows
		id			= server.nextNodeID;
		
		msg			= List[ '/local', id, '[', '/new', className ];
		this.prAddArgs( msg, args );
		msg.add( ']' );
		
//		server.sendBundle( nil, [ '/local', id, '[', '/new', className ] ++ args ++ [ ']' ]);
//msg.postln;
		server.listSendMsg( msg );
	}
	
	destroy {
		server.sendMsg( '/free', id );
	}
	
	*destroyAll {
		var list;
		list = allObjects.copy;
		allObjects = Array.new( 8 );
		list.do({ arg obj; obj.destroy; });
	}

	doesNotUnderstand { arg selector ... args;
		server.listSendMsg( this.prMethodCall( nil, selector, args ));
	}
	
	prAddArgs { arg list, args;
		args.do({ arg x;
			if( x.respondsTo( \id ), {
				list.addAll([ '[', '/ref', x.id, ']' ]);
			}, {
				list.addAll( x.asSwingArg );
			});
		});
	}

	prMethodCall { arg list, selector, args;
		list = list ?? { List.new; };
		list.add( '/method' );
		list.add( id );
		list.add( selector );
		this.prAddArgs( list, args );
		^list;
	}
	
	*newFrom { arg javaObject, selector ... args;
		^super.new.prInitJavaResult( javaObject, selector, args );
	}
	
	prInitJavaResult { arg javaObject, selector, args;
		var msg;
		
		server		= javaObject.server;
		allObjects	= allObjects.add( this );	// array grows
		id			= server.nextNodeID;
		msg			= List[ '/local', id, '[' ];
		javaObject.prMethodCall( msg, selector, args );
		msg.add( ']' );
		server.listSendMsg( msg );
	}
}